%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%                                                                         %%
%%                        Welcome to GENFIRE!                              %%
%%           GENeralized Fourier Iterative REconstruction                  %%
%%                                                                         %%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%% GENFIRE is a robust, Fourier-based reconstruction algorithm that is
%% capable of using a limited set of input projections to generate a 3D reconstruction
%% while also partially retrieving missing projection information. It does this by iterating 
%% between real and reciprocal space and applying simple constraints in each to find an optimal solution that  
%% correlates with the input projections while simultaneously obeying real space conditions
%% such as positivity (the assumption that there is no negative electron density)
%% and support (we require that the 3D object resulting from zero padded projections
%% exists entirely within some smaller region). The result is a more consistent 
%% and faithful reconstruction with superior contrast and, in many cases, resolution when
%% compared with more traditional 3D reconstruction algorithms such as
%% Filtered Back-Projection. In addition, GENFIRE is capable of dynamically identifying
%% the noise level of the input projections (out to what resolution does useful information exist
%% before noise dominates the signal) without any prior knowledge from the user. 

%% Author: AJ Pryor, Yongsoo Yang
%% Jianwei (John) Miao Coherent Imaging Group
%% University of California, Los Angeles
%% Copyright (c) 2015. All Rights Reserved.

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%




%%%   User Parameters   %%%

filename_Projections = 'FePt_Denoised_Projections.mat'; %%filename of projections, which should be size NxNxN_projections where N_projections is the number of projections

filename_Angles = 'FePt_Initial_Angles.mat';%%angles can be either a 1xN_projections array containing a single tilt series, or
%%a 3xN_projections array containing 3 Euler angles for each projections in the form [phi;theta;psi]

%filename_Support = './models/betagal/support.mat'; %% NxNxN binary array specifying a region of 1's in which the reconstruction can exist 

%filename_InitialModel = '';

ReusltFileName = 'Phi_refined_angle.mat';

anglesearch_phi = -2.1:0.3:2.1;


% size of input projection array
nn=256;

global numIterations 
numIterations = 100; 

global pixelSize
pixelSize = 4; 

GriddingMethod = 2; %Gridding method 1: FFT gridding with C mexa function (fastest)
                    %Gridding method 2: FFT gridding with MATLAB function (slower than C mexa function)
                    %Gridding method 3: DFT gridding (much slower than FFT, but the most accurate)

oversamplingRatio =4; %%The code will zero-pad projections for you to the inputted oversampling ratio. If your projections are already oversampled
%%then set this to 1.

interpolationCutoffDistance =.70; %%radius of sphere (in pixels) within which to include measured datapoints 
%%when assembling the 3D Fourier grid


%% If you do not need FSC, set ComputeFourierShellCorrelation to 0 for speed as the FSC calculation requires reconstructing everything twice


complexCutoff = 1; %value between 0 and 1 indicating percentage of maximum resolution to enforce as complex values
magnitudeCutoff = .0;%value between 0 and 1 indicating percentage of maximum resolution to enforce as magnitude only
%%%%****IN CURRENT IMPLEMENTATION ONLY COMPLEX VALUES ARE ENFORCED
%%%%****FUTURE IMPLEMENTATIONS WILL CONTAIN PHASE RETRIEVAL FOR PARTIAL
%%%%****MAGNITUDES


%%%   Begin Reconstruction   %%%
clc
angles = single(importdata(filename_Angles));%casting to single saves memory
projections = single(importdata(filename_Projections));
projections_ori = single(importdata(filename_Projections)); %original projection without padding


% check if initial model is provided
if exist('filename_InitialModel','var')
    initialObject = single(importdata(filename_InitialModel));
else
    initialObject = [];
end


global support %make support variable globally accessable to avoid passing copies of large arrays around to different functions
if exist('filename_Support','var')
    support = single(importdata(filename_Support));
else
    support = single(ones(nn,nn,nn));
end



%zero pad projections to user-inputted oversampling ratio
numProj = size(projections,3);
dim = size(projections,2);%original array size
newDim = dim*oversamplingRatio;%size of oversampled array
padding = (newDim-dim)/2;%how many zeros to add
nc = newDim/2+1;%central pixel of larger array
n2 = dim/2;%radius of smaller array
projections = padarray(projections,[padding padding 0]);%zero pad projections
support = padarray(support,[padding padding padding]); %%zero pad support

%If there is only one tilt angle (such as in tomography), fill in other
%Euler angles with zeros
if size(angles,1) ==1
angles = [angles;zeros(1,length(angles));zeros(1,length(angles))];
end





%construct normalized spatial frequency indices
ncK = size(projections,2)/2+1;%central pixel
n2K = ncK-1;%max radius
[Kx Ky Kz] = meshgrid(-n2K:n2K-1,-n2K:n2K-1,-n2K:n2K-1);%grid of Fourier indices
KMags = sqrt(Kx.^2+Ky.^2+Kz.^2);%take magnitude
KMags = KMags./n2K; %normalized Fourier indices for defining resolution



fprintf('GENFIRE: Assembling Fourier grid...\n\n');

%%INTERPOLATE TO GRID
%%fillInFourierGrid_C contains the c++ code "weightVals.cpp" that must be
%%compiled on your local machine to run. If you cannot get it to work with
%%the command "mex weightVals.cpp" then use the version fillInFourierGrid
%%which does the same thing but slower.
tic
if GriddingMethod == 1
   [recIFFT measuredK] = fillInFourierGrid_C(projections,angles,interpolationCutoffDistance);%interpolate projections to Fourier grid
elseif GriddingMethod == 2
   [recIFFT measuredK] = fillInFourierGrid(projections,angles,interpolationCutoffDistance);%interpolate projections to Fourier grid
elseif GriddingMethod == 3
      % gridding with DFT (using My_fill_cubicgrid_ver1), this is much slower than FFT interpolation, but more accurate
    measuredK = double(My_fill_cubicgrid_ver1(dim, dim, ori_projections, angles, interpolationCutoffDistance, 2, newDim, newDim));
    recIFFT = real(my_ifft(measuredK));
end
toc

if complexCutoff == 1;
    complexCutoff = 1.5; %% if all of the information is enforceable, I choose to also enforce super resolution as simulations 
    %% indicated doing so produced better reconstructions
end
constraintInd_complex = find(measuredK~=0&KMags<=complexCutoff);%indices of measured Fourier points to enforce with complex values
constraintInd_magnitude = find(measuredK~=0&KMags<=magnitudeCutoff);
% constraintInd_magnitude = [];%This is a null placeholder for future versions


% recIFFT=smooth3D(recIFFT,complexCutoff);%To low pass filter the IFFT reconstruction (which will be our initial
%%object if none was inputted) to the resolution that we can enforce.

fprintf('GENFIRE: Reconstructing... \n\n');
if isempty(initialObject) 
    [GENFIRE_rec, errK] = GENFIRE_iterate(numIterations,recIFFT,support,measuredK,constraintInd_complex,constraintInd_magnitude,[],[]);%run iterations
else
    [GENFIRE_rec, errK] = GENFIRE_iterate(numIterations,initialObject,support,measuredK,constraintInd_complex,constraintInd_magnitude,[],[]);
end
reconstructionTime = toc;
reconstructionTime = round(10*reconstructionTime)./10;
fprintf('GENFIRE: Reconstruction completed in %.12g seconds\n\n',reconstructionTime);

GENFIRE_rec = GENFIRE_rec(ncK-n2:ncK+n2-1,ncK-n2:ncK+n2-1,ncK-n2:ncK+n2-1);%extract the original sized result from the oversampled reconstruction


%%

new_angles = angles;

% initialize array for storing error
Errar = zeros(size(projections,3),length(anglesearch_phi),3);

% padd some zeros to reconstructed volume prevent artifact during forward-projection
pph = 20;
GENar_pad = zeros(nn+pph*2,nn+pph*2,nn+pph*2);
GENar_pad(pph+1:nn+pph,pph+1:nn+pph,pph+1:+nn+pph) = GENFIRE_rec;

    
% loop over all projections  
for i=1:size(angles,2)     
    
    % loop over the angle to be searched
    for j=1:length(anglesearch_phi)
        % calculate forward projection to the given angle using
        % interpolation
        proj = My_get_3D_projection_interp(GENar_pad,angles(1,i)+anglesearch_phi(j), ...
                                               angles(2,i),angles(3,i),50);                
        proj = proj(pph+1:nn+pph,pph+1:nn+pph);
        
        % compare the calculated projection with measured projection and
        % calculate error
        Errar(i,j,2) = My_proj_compare(squeeze(projections_ori(:,:,i)),proj);
    end
       
    % determine the angle which gives minimum error
    [~,minerrind] = min(squeeze(Errar(i,:,1)));
    new_angles(1,i) = angles(1,i)+anglesearch_phi(minerrind);
end
    

save(ReusltFileName, 'new_angles');

