% My_fill_cubicgrid
% Y. Yang, UCLA Physics & Astronomy
% First version date: 2015. 04. 30.
% output parameter: ftArray (n x n x nz) Fourier 3D array
% input parameter: n_ori (original projection 2nd dim size)   
%                  nz_ori (original projection 1st dim size)                  
%                  PROJvol - measured projection, (nz x n x P) array with P # of projections
%                  angles - 3 Euler angle array, (3 x P) array
%                  K_thresh - threshold for acceptible distance
%                  Typeind - 1 for picking one minimal distance value
%                            2 for weighted averaging all values within threshold
%                   n (length of oversampled projection 2nd dim)
%                    nz (length of oversampled projection 1st dim) 

function ftArray = My_fill_cubicgrid_ver1(n_ori, nz_ori, PROJvol, angles,K_thresh,Typeind, n, nz)

% if distance below minInvThresh, minInvThresh will be used
% this is to prevent division by zero
minInvThresh = 0.001;

% initialize normal vectors and rotation matrices array
normVECs = zeros(size(PROJvol,3),3);
rotMATs = zeros(3,3,size(PROJvol,3));

phis = angles(1,:);
thetas = angles(2,:);
psis = angles(3,:);

% calculate rotation matrices and normal vectors from the rotation matrices
for i=1:size(PROJvol,3)
    vector1 = [0 0 1];
    rotmat1 = MatrixQuaternionRot(vector1,phis(i));
    
    vector2 = [0 1 0];
    rotmat2 = MatrixQuaternionRot(vector2,thetas(i));

    vector3 = [0 0 1];
    rotmat3 = MatrixQuaternionRot(vector3,psis(i));

    rotMATs(:,:,i) =  rotmat3*rotmat2*rotmat1;

    init_normvec = [0 1 0];
    normVECs(i,:) = squeeze(rotMATs(:,:,i))*init_normvec';
end

   
% initiate Fourier space indices
if mod(n,2) == 1
    kx = -1*(n-1)/2:1:(n-1)/2;
    kx_ext = kx;
else
    kx_ext = -1*n/2:1:n/2;
end

% initiate Fourier space indices
if mod(nz,2) == 1
    kz_half = -1*(nz-1)/2:1:0;
    nz_half = (nz-1)/2;
else
    kz_half = -1*nz/2:1:0;
    nz_half = nz/2;
end

ky_ext = kx_ext;

% Fourier grid
% To save time, only half of kz will be interpolated
% and centrosymmetricity will be enforced later
%[KX, KY, KZ] = meshgrid(kx,ky,kz_half);
[KY, KX, KZ] = meshgrid(ky_ext,kx_ext,kz_half);
    
% initialize variables
FS = single(zeros(size(KX))*(-1)); % Fourier points
Numpt = single(zeros(size(KX))); % array to store how many points found per Fourier point
Mindist = single(ones(size(KX))*(10000)); % array to store current minimum distance
invSumdist = single(zeros(size(KX))); % array to store sum of inverse distance of Fourier point

% initiate Fourier space indices
if mod(n_ori,2) == 1
    kx_ori = -1*(n_ori-1)/2:1:(n_ori-1)/2;
else
    kx_ori = -1*n_ori/2:1:n_ori/2-1;
end

% initiate Fourier space indices
if mod(nz_ori,2) == 1
    kz_ori = -1*(nz_ori-1)/2:1:(nz_ori-1)/2;    
else
    kz_ori = -1*nz_ori/2:1:nz_ori/2-1;
end
    
%[KX0, KY0, KZ0] = meshgrid(0,ky,kz);
[KY0, KZ0] = meshgrid(kx_ori,kz_ori);
    
for p=1:size(PROJvol,3)
    %p
    % current projection
    curr_proj = squeeze(PROJvol(:,:,p));
    
    [KY, KX, KZ] = meshgrid(ky_ext,kx_ext,kz_half);
    
    % find Fourier points within the threshold
    D = distancePointsPlane_YY([KX(:) KY(:) KZ(:)]', normVECs(p,:));
    Dind = find(D < K_thresh);
    
    % rotate the plane to zero degree
    CP = closestpoint(normVECs(p,:)',0,[KX(Dind); KY(Dind); KZ(Dind)]);
    CP_plane = (squeeze(rotMATs(:,:,p)))\CP;
    
    %clear KX KY KZ
    % picked closest point which is within the projection plain, x coordinate must be zero after rotation
    if sum(abs(CP_plane(2,:)) > 0.0001) > 0
        fprintf(1,'something wrong!\n');                
    end
    
    % consider Fourier points only within the resolution circle
    Gind = Dind(abs(CP_plane(1,:)) <= n/2 & abs(CP_plane(3,:)) <= n/2);  % good indices  
    G_CP_plane = CP_plane(:,abs(CP_plane(1,:)) <= n/2 & abs(CP_plane(3,:)) <= n/2);  % good in-plane coordinates
    
    % determine the available memory in MB
    if ispc % in case of Windows machine
        [~, SYSTEMVIEW] = memory;
        av_memory_size = SYSTEMVIEW.PhysicalMemory.Available / 1000000;
    elseif isunix && ~ismac  % in case of linux (or unix)
        [~,out]=system('cat /proc/meminfo | grep MemFree');
        av_memory_size=sscanf(out,'MemFree:          %d kB');
        av_memory_size = av_memory_size / 1000;
    else % in case of mac (I don't have mac now, to be implemented later)
        av_memory_size = 1000;
    end
    
    memory_per_index = 40*length(curr_proj(:))/1000000;        
    
    % determine block size for vectorized calculation
    block_size = floor(av_memory_size/memory_per_index);
    if block_size < 1
        block_size = 1;
    end
    %block_size = 500;
    cutnum = floor(length(Gind)/block_size);
    cutrem = mod(length(Gind),block_size);
    
    % loop over Fourier points within the threshold
    for i=1:cutnum        
        curr_indices = ((i-1)*block_size+1):(i*block_size);
        
        % DFT calculation
        Fpoints = sum(bsxfun(@times, curr_proj(:), exp(-1*1i*2*pi*(KY0(:)*G_CP_plane(1,curr_indices)/n+KZ0(:)*G_CP_plane(3,curr_indices)/n))),1);
        
        if Typeind == 1 % minimum distance point
            Minind = find(D(Gind(curr_indices)) < Mindist(Gind(curr_indices)));
            FS(Gind(curr_indices(Minind))) = Fpoints(Minind);
            Mindist(Gind(curr_indices(Minind))) = D(Gind(curr_indices(Minind)));
        elseif Typeind == 2  %weighted avearaging
            currDist = D(Gind(curr_indices));
            currDist(currDist < minInvThresh) = minInvThresh; % if distance smaller than minInvThresh, put minInvThresh (to prevent divison by zero)
            currInvDist = 1./ currDist;           % inverse distance
            
            % re-average inverse distance
            FS(Gind(curr_indices)) = FS(Gind(curr_indices)).* invSumdist(Gind(curr_indices)) + currInvDist.*Fpoints;
            invSumdist(Gind(curr_indices)) = invSumdist(Gind(curr_indices)) + currInvDist;
            FS(Gind(curr_indices)) = FS(Gind(curr_indices)) ./ invSumdist(Gind(curr_indices));
            Numpt(Gind(curr_indices)) = Numpt(Gind(curr_indices)) + 1;
        end
            
        clear Fpoints
    end
    
    % do the same thing for remaning block after the loop
    if cutrem~=0
        curr_indices = (cutnum*block_size+1):length(Gind);
        Fpoints = sum(bsxfun(@times, curr_proj(:), exp(-1*1i*2*pi*(KY0(:)*G_CP_plane(1,curr_indices)/n+KZ0(:)*G_CP_plane(3,curr_indices)/n))),1);
        
        if Typeind == 1
            Minind = find(D(Gind(curr_indices)) < Mindist(Gind(curr_indices)));
            FS(Gind(curr_indices(Minind))) = Fpoints(Minind);
            Mindist(Gind(curr_indices(Minind))) = D(Gind(curr_indices(Minind)));
        elseif Typeind == 2
            currDist = D(Gind(curr_indices));
            currDist(currDist < minInvThresh) = minInvThresh;
            currInvDist = 1./ currDist;           
            FS(Gind(curr_indices)) = FS(Gind(curr_indices)).* invSumdist(Gind(curr_indices)) + currInvDist.*Fpoints;
            invSumdist(Gind(curr_indices)) = invSumdist(Gind(curr_indices)) + currInvDist;
            FS(Gind(curr_indices)) = FS(Gind(curr_indices)) ./ invSumdist(Gind(curr_indices));
            Numpt(Gind(curr_indices)) = Numpt(Gind(curr_indices)) + 1;

        end
        
        clear Fpoints
    end
end

clear Mindist invSumdist Numpt

% enforce centrosymmetricity
ftArray = single(zeros(n,n,nz));

if mod(n,2) == 1
    ftArray_half = reshape(FS,n,n,nz_half+1);
    ftArray(:,:,1:nz_half+1) = ftArray_half;
    
    if mod(nz,2) == 1
        ftArray_ahalf = ftArray_half(:,:,1:end-1);
    else
        ftArray_ahalf = ftArray_half(:,:,2:end-1);
    end
    
    ftArray_ahalf = rot90(ftArray_ahalf, 2);    
    ftArray_ahalf = permute(ftArray_ahalf, [3 1 2]);
    ftArray_ahalf = flipud(ftArray_ahalf);
    ftArray_ahalf = permute(ftArray_ahalf, [2 3 1]);
    
    ftArray(:,:,nz_half+2:end) = conj(ftArray_ahalf);
else
    ftArray_half_ext = reshape(FS,n+1,n+1,nz_half+1);
    ftArray_half = ftArray_half_ext(1:end-1,1:end-1,:);
    ftArray(:,:,1:nz_half+1) = ftArray_half;
    
    if mod(nz,2) == 1
        ftArray_ahalf_ext = ftArray_half_ext(:,:,1:end-1);
    else
        ftArray_ahalf_ext = ftArray_half_ext(:,:,2:end-1);   
    end
    ftArray_ahalf_ext = rot90(ftArray_ahalf_ext, 2);
    ftArray_ahalf_ext = permute(ftArray_ahalf_ext, [3 1 2]);
    ftArray_ahalf_ext = flipud(ftArray_ahalf_ext);
    ftArray_ahalf_ext = permute(ftArray_ahalf_ext, [2 3 1]);
    
    ftArray(1:end,1:end,nz_half+2:end) = conj(ftArray_ahalf_ext(1:end-1,1:end-1,:));
    
end

end



function D = distancePointsPlane_YY(points, normvec)
    %distancePointsPlane_YY unsigned distances betwen 3D points and a plane
    % through origin
    %
    %   D = distancePointsPlane_YY(point, normvec)
    %   Returns the euclidean distance between points and a plane going through origin with normal vector normvec,
    %   given by: 
    %   points : (3 x n) array of 3D vectors
    %   normvec : (3 x 1) or (1 x 3) array of normal vector
    %   D     : (1 x n) vector  
 
    %
    %   ---------
    %   author : Y. Yang, UCLA Physics and Astronomy 
    %   created the 05/03/2015.
    %

    % normalized plane normal
    normvec = normvec(:) / norm(normvec);
    D = abs(points(1,:)*normvec(1) + points(2,:)*normvec(2) + points(3,:)*normvec(3));

end


function x = closestpoint(n, d, p)
    % n is the vector [A,B,C] that defines the plane
    % d is the distance of the plane from the origin
    % p is the point  [P,Q,R]
    if size(p,2) == 1
        v = (d - sum(p.*n)) / sum(n.*n);
        x = p + v * n;
    else
        nr = repmat(n,[1 size(p,2)]);
        v = (d - sum(p.*nr,1)) / sum(n.*n);
        x = p + repmat(v,[3 1]) .* nr;
    end
end

function dd = MatrixQuaternionRot(vector,theta)

theta = theta*pi/180;
vector = vector/sqrt(dot(vector,vector));
w = cos(theta/2); x = -sin(theta/2)*vector(1); y = -sin(theta/2)*vector(2); z = -sin(theta/2)*vector(3);
RotM = [1-2*y^2-2*z^2 2*x*y+2*w*z 2*x*z-2*w*y;
      2*x*y-2*w*z 1-2*x^2-2*z^2 2*y*z+2*w*x;
      2*x*z+2*w*y 2*y*z-2*w*x 1-2*x^2-2*y^2;];

dd = RotM;
end


