%%  fillInFourierGrid %%

%%inputs:
%%  projections - measured projections
%%  angles - Euler angles in the form 3xN_projections, where each projection has 3 angles in the form [phi;theta;psi]
%%  interpolationCutoffDistance - radius of sphere in which to include measured
%%      values when filling in grid. All points within this sphere will be weighted
%%      linearly by their inverse distance. 

%%outputs:
%%  rec - inverse FFT of the assembled Fourier grid
%%  measuredK -assembled Fourier Grid

%% Author: AJ Pryor
%% Jianwei (John) Miao Coherent Imaging Group
%% University of California, Los Angeles
%% Copyright (c) 2015. All Rights Reserved.




function [rec measuredK] = fillInFourierGrid(projections,angles,interpolationCutoffDistance)
tic

dim1 = size(projections,1);
dim2 = size(projections,2);
nc = single(round((dim1+1)/2));%center pixel
n2 = single(nc-1);%radius of array

%initialize arrays to contain coordinates
measuredY = zeros(1,size(projections,2)*size(projections,1),size(projections,3),'single');
measuredX = zeros(1,size(projections,2)*size(projections,1),size(projections,3),'single');
measuredZ = zeros(1,size(projections,2)*size(projections,1),size(projections,3),'single');

%initialize array for measured Fourier points
kMeasured = zeros(dim1,dim1,size(projections,3),'single');

for projNum = 1:size(projections,3);
phi = angles(1,projNum);
theta = angles(2,projNum);
psi = angles(3,projNum);

%rotation matrix (by preference it operates on vectors [y;x;z] NOT [x;y;z].
%I prefer this way because in MATLAB the first index corresponds to the row
%which I find more natural to consider the y-axis
R = [ -sind(psi)*sind(phi)+cosd(theta)*cosd(phi)*cosd(psi)  -sind(psi)*cosd(phi)-cosd(theta)*sind(phi)*cosd(psi) cosd(psi)*sind(theta) ;...
      cosd(psi)*sind(phi)+cosd(theta)*cosd(phi)*sind(psi) cosd(psi)*cosd(phi)-cosd(theta)*sind(phi)*sind(psi) sind(psi)*sind(theta); ...
    -sind(theta)*cosd(phi) sind(theta)*sind(phi) cosd(theta)];
[ky kz] = meshgrid(-n2:n2-1,-n2:n2-1);kz = single(kz(:))'; ky = single(ky(:))'; %initialize coordinates of unrotate projection slice
kx = zeros(1,dim1*dim1,'single'); %0 degree rotation is a projection onto the Y-Z plane, so all points have kx=0;

rotkCoords = R*[ky;kx;kz];%rotate coordinates

measuredY(:,:,projNum) = rotkCoords(1,:);%rotated Y
measuredX(:,:,projNum) = rotkCoords(2,:);%rotated X
measuredZ(:,:,projNum) = rotkCoords(3,:);%rotated Z
kMeasured(:,:,projNum) = my_fft(projections(:,:,projNum)); %track complex value at that point
end
clear projections

%reshape to simplify
measuredX = reshape(measuredX,1,size(kMeasured,2)*size(kMeasured,1)*size(kMeasured,3));
measuredY = reshape(measuredY,1,size(kMeasured,2)*size(kMeasured,1)*size(kMeasured,3));
measuredZ = reshape(measuredZ,1,size(kMeasured,2)*size(kMeasured,1)*size(kMeasured,3));
kMeasured = reshape(kMeasured,1,size(kMeasured,2)*size(kMeasured,1)*size(kMeasured,3));

masterInd = [];%masterInd will be a large list of the grid indices
masterVals = [];%complex values to include in weighted averaging for those grid points
masterDistances = [];%distance from measured value to grid point

shiftMax = round(interpolationCutoffDistance);
% shiftMax = 0;
%The nearest grid point to a measured value can be found by rounding, but
%there can be more than one grid point within the cutoff sphere, so must
%search locally for other possibilities. However in practice I have found 
%this search can slow the program down greatly, without significant change
%in final result. Even searching 1 voxel in either direction increases the
%number of calculations by 3^3 = 27; For this reason I often set shiftMax = 0 and
%just assign values to their closest voxel.
   for Yshift = -shiftMax:shiftMax 
       for Xshift = -shiftMax:shiftMax
           for Zshift = -shiftMax:shiftMax
                tmpX = (round(measuredX)+Xshift); % apply shift
                tmpY = (round(measuredY)+Yshift);
                tmpZ = (round(measuredZ)+Zshift);
                tmpVals = kMeasured;
                distances = sqrt(abs(measuredX-tmpX).^2+abs(measuredY-tmpY).^2+abs(measuredZ-tmpZ).^2); 
                tooFarInd = find(distances>interpolationCutoffDistance); %throw out values outside of interpolation cutoff distance
                tmpY = tmpY+nc;
                tmpZ = tmpZ+nc;
                tmpX = tmpX+nc;
                outsideResolutionInd = find(tmpX>dim1|tmpX<1|tmpY>dim1|tmpY<1|tmpZ>dim1|tmpZ<1); %throw out values outside of our cubic volume
                badIndices = union(outsideResolutionInd,tooFarInd); %combine bad indices and delete them
                tmpY(badIndices) = [];
                tmpX(badIndices) = [];
                tmpZ(badIndices) = [];
                distances(badIndices) = [];
                tmpVals(badIndices) = [];
                masterInd = [masterInd sub2ind([dim1 dim1 dim1],tmpY,tmpX,tmpZ)]; %append values to lists
                masterVals = [masterVals tmpVals];
                masterDistances = [masterDistances distances];
           end
       end
   end

[masterInd sortInd] = sort(masterInd);%organize lists
masterVals = masterVals(sortInd);
masterDistances = masterDistances(sortInd);

measuredK = zeros(dim1,dim1,dim1,'single');


[uniqueVals uniqueInd] = unique(masterInd);%find non repeating values

uniqueInd(end+1) = length(masterInd)+1;
%%There is potential for a very large increase in speed if this loop can be
%%vectorized

diffVec = diff(uniqueInd);
singleInd = find(diffVec==1);
multiInd = find(diffVec~=1);
measuredK(uniqueVals(singleInd)) = masterVals(uniqueInd(singleInd));

% for index = 1:length(multiInd)-1
%     ind = multiInd(index);
%     distances = masterDistances(uniqueInd(ind):uniqueInd(ind+1)-1)+1e-10;
%     complexVals = masterVals(uniqueInd(ind):uniqueInd(ind+1)-1);
%     weights = ((1./distances).*sum(distances)./sum((1./distances).*sum(distances)));
%     voxel_Mag = sum(weights.*abs(complexVals));
%     voxel_Phase = angle(sum(weights.*complexVals));
%     measuredK(masterInd(uniqueInd(ind))) = voxel_Mag.*exp(1i.*voxel_Phase);
%   
% end
[weightedMagnitudes Rv Cv] = weightVals(multiInd',double(masterDistances),double(masterVals),double(uniqueInd));
weightedPhases = angle((Rv+i*Cv));
measuredK(masterInd(uniqueInd(multiInd))) = weightedMagnitudes.*exp(1*i*weightedPhases);

rec = real(my_ifft(measuredK));
timeTakenToFillInGrid = toc;
timeTakenToFillInGrid = round(10*timeTakenToFillInGrid)./10;
fprintf('GENFIRE: Fourier grid assembled in %.12g seconds\n\n',timeTakenToFillInGrid);
