%%  GENFIRE_iterate %%

%%inputs:
%%  numIterations - number of iterations to run
%%  support - region of 1's and 0's defining where the reconstruction can exist (voxels that are 0 in support will be set to 0 in reconstruction)
%%  measuredK - measured Fourier points
%%  constraintInd_complex - indices of measuredK that can be enforced as complex (magnitude and phase) values
%%  constraintInd_magnitude - indices of measuredK that can be enforced as magnitude values
%%  R_freeInd_complex - indices of 5% of points in the highest resolution shell of measuredK that are being withheld from 
%%          reconstruction and compared to after iteration. The reconstructability of these points may be used to determine to 
%%          what resolution the input projections are correlated enough to generate a globally consistent reconstruction.
%%  R_freeInd_mag - same as R_freeInd_complex, except these indices correspond to measuredK voxels being withheld that can
%%          only be enforced as magnitudes.

%%outputs:
%%  rec - reconstruction after iteration
%%  errK - reciprocal space error
%%  Rfree_complex - value of complex Rfree vs iteration
%%  Rfree_magnitude - value of magnitude Rfree vs iteration

%% Author: AJ Pryor
%% Jianwei (John) Miao Coherent Imaging Group
%% University of California, Los Angeles
%% Copyright (c) 2015. All Rights Reserved.

function [rec, errK, Rfree_complex, Rfree_magnitude] = GENFIRE_iterate(numIterations,initialObject,support,measuredK,constraintInd_complex,constraintInd_magnitude,R_freeInd_complex,R_freeInd_mag)
bestErr = 1e30;%initialize best error
Rfree_complex = -1*ones(1,numIterations,'single');%% initialize Rfree_complex curve , -1 is a flag that means undefined
Rfree_magnitude = -1*ones(1,numIterations,'single');%% initialize Rfree_magnitude curve
errK = zeros(1,numIterations,'single');
combinedConstraintInd = union(constraintInd_complex,constraintInd_magnitude);%combined indices of voxels that are enforced in any way
for iterationNum = 1:numIterations
% if mod(iterationNum,10)==0
%     iterationNum
% end
initialObject(initialObject<0) = 0; %enforce positivity
initialObject = initialObject.*support;%enforce support
k = my_fft(initialObject);%take FFT of initial object

%monitor error
% errK(iterationNum) =sum(abs(k(combinedConstraintInd)-measuredK(combinedConstraintInd)))./sum(abs(measuredK(combinedConstraintInd)));%complex
errK(iterationNum) = sum(abs(abs(k(combinedConstraintInd))-abs(measuredK(combinedConstraintInd))))./sum(abs(measuredK(combinedConstraintInd)));%magnitudes only

if nargin > 6 %if values have been withheld from measuredK for monitoring R_free, check them accordingly
    if ~isempty(R_freeInd_complex)
        Rfree_complex(iterationNum) = sum(abs(k(R_freeInd_complex)-measuredK(R_freeInd_complex)))./sum(abs(measuredK(R_freeInd_complex)));
    end
    if ~isempty(R_freeInd_mag)
        Rfree_magnitude(iterationNum) = sum(abs(abs(k(R_freeInd_mag))-abs(measuredK(R_freeInd_mag))))./sum(abs(measuredK(R_freeInd_mag)));
    end
end
if errK(iterationNum)<bestErr %if current reconstruction has better error, update best error and best reconstruction
    bestErr = errK(iterationNum);
    rec = initialObject;
end

%%the below magnitude enforcement does very little to improve the reconstructions
%%because there is no phase information. Will be improved in future
%%releases
if iterationNum == 1; %for magnitude enforcement, provide random phases for first iteration
    randPhases = single(rand(size(k))*2*pi);
    k(constraintInd_magnitude) = abs(measuredK(constraintInd_magnitude)).*exp(1*i*randPhases(constraintInd_magnitude));
    bestErr = 1e30;%initialize best error
else
    %replace magnitude of measured value with phase of current value
    k(constraintInd_magnitude) = abs(measuredK(constraintInd_magnitude)).*exp(1i*angle(k(constraintInd_magnitude)));
end
k(constraintInd_complex) = measuredK(constraintInd_complex);%replace complex measured values
initialObject = real(my_ifft(k));%obtain next object with IFFT
end
