%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%                                                                         %%
%%                        Welcome to GENFIRE!                              %%
%%           GENeralized Fourier Iterative REconstruction                  %%
%%                                                                         %%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%% GENFIRE is a robust, Fourier-based reconstruction algorithm that is
%% capable of using a limited set of input projections to generate a 3D reconstruction
%% while also partially retrieving missing projection information. It does this by iterating 
%% between real and reciprocal space and applying simple constraints in each to find an optimal solution that  
%% correlates with the input projections while simultaneously obeying real space conditions
%% such as positivity (the assumption that there is no negative electron density)
%% and support (we require that the 3D object resulting from zero padded projections
%% exists entirely within some smaller region). The result is a more consistent 
%% and faithful reconstruction with superior contrast and, in many cases, resolution when
%% compared with more traditional 3D reconstruction algorithms such as
%% Filtered Back-Projection. In addition, GENFIRE is capable of dynamically identifying
%% the noise level of the input projections (out to what resolution does useful information exist
%% before noise dominates the signal) without any prior knowledge from the user. 

%% Author: AJ Pryor and Yongsoo Yang
%% Jianwei (John) Miao Coherent Imaging Group
%% University of California, Los Angeles
%% Copyright (c) 2015. All Rights Reserved.

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%




%%%   User Parameters   %%%

filename_Projections = './data/test_projections.mat'; %%filename of projections, which should be size NxNxN_projections where N_projections is the number of projections

filename_Angles = './data/test_angles.mat';%%angles can be either a 1xN_projections array containing a single tilt series, or
%%a 3xN_projections array containing 3 Euler angles for each projections in the form [phi;theta;psi]

filename_Support = './data/test_support.mat'; %% a binary array specifying a region of 1's in which the reconstruction can exist 

%filename_InitialModel = '';

filename_results = './results/test_reconstruction_result.mat';

% input projection size
nn=64;

global numIterations 
numIterations = 50; 

GriddingMethod = 2; %Gridding method 1: FFT gridding with C mexa function (fastest) (for detais about mex function complile,
                    %please see "MEXA_fuction_install_notes.txt" included in this package).
                    %Gridding method 2: FFT gridding with MATLAB function (slower than C mexa function)
                    %Gridding method 3: DFT gridding (much slower than FFT, but the most accurate)

oversamplingRatio =3; %%The code will zero-pad projections for you to the inputted oversampling ratio. If your projections are already oversampled
%%then set this to 1.

interpolationCutoffDistance =.7; %%radius of sphere (in pixels) within which to include measured datapoints 


%% If you do not need FSC, set ComputeFourierShellCorrelation to 0 for speed as the FSC calculation requires reconstructing everything twice


complexCutoff = 1; %value between 0 and 1 indicating percentage of maximum resolution to enforce as complex values
magnitudeCutoff = .0;%value between 0 and 1 indicating percentage of maximum resolution to enforce as magnitude only
%%%%****IN CURRENT IMPLEMENTATION ONLY COMPLEX VALUES ARE ENFORCED
%%%%****FUTURE IMPLEMENTATIONS WILL CONTAIN PHASE RETRIEVAL FOR PARTIAL
%%%%****MAGNITUDES


%%%   Begin Reconstruction   %%%
clc
angles = single(importdata(filename_Angles));%casting to single saves memory
projections = single(importdata(filename_Projections));
ori_projections = single(importdata(filename_Projections));


if exist('filename_InitialModel','var')
    initialObject = single(importdata(filename_InitialModel));
else
    initialObject = [];
end


global support %make support variable globally accessable to avoid passing copies of large arrays around to different functions
if exist('filename_Support','var')
    support = single(importdata(filename_Support));
else
    support = single(ones(nn,nn,nn));
end


%zero pad projections to user-inputted oversampling ratio
numProj = size(projections,3);
dim = size(projections,2);%original array size
newDim = dim*oversamplingRatio;%size of oversampled array
padding = (newDim-dim)/2;%how many zeros to add
nc = newDim/2+1;%central pixel of larger array
n2 = dim/2;%radius of smaller array
projections = padarray(projections,[padding padding 0]);%zero pad projections
%support = padarray(support,[padding padding padding]); %%zero pad support
support = My_paddzero(support,[newDim newDim newDim]);

%If there is only one tilt angle (such as in tomography), fill in other
%Euler angles with zeros
if size(angles,1) ==1
angles = [angles;zeros(1,length(angles));zeros(1,length(angles))];
end

%construct normalized spatial frequency indices
ncK = size(projections,2)/2+1;%central pixel
n2K = ncK-1;%max radius
[Kx Ky Kz] = meshgrid(-n2K:n2K-1,-n2K:n2K-1,-n2K:n2K-1);%grid of Fourier indices
KMags = sqrt(Kx.^2+Ky.^2+Kz.^2);%take magnitude
KMags = KMags./n2K; %normalized Fourier indices for defining resolution


fprintf('GENFIRE: Assembling Fourier grid...\n\n');

%%INTERPOLATE TO GRID
%%fillInFourierGrid_C contains the c++ code "weightVals.cpp" that must be
%%compiled on your local machine to run. If you cannot get it to work with
%%the command "mex weightVals.cpp" then use the version fillInFourierGrid
%%which does the same thing but slower.
tic
if GriddingMethod == 1
   [recIFFT measuredK] = fillInFourierGrid_C(projections,angles,interpolationCutoffDistance);%interpolate projections to Fourier grid
elseif GriddingMethod == 2
   [recIFFT measuredK] = fillInFourierGrid(projections,angles,interpolationCutoffDistance);%interpolate projections to Fourier grid
elseif GriddingMethod == 3
      % gridding with DFT (using My_fill_cubicgrid_ver1), this is much slower than FFT interpolation, but more accurate
    measuredK = double(My_fill_cubicgrid_ver1(dim, dim, ori_projections, angles, interpolationCutoffDistance, 2, newDim, newDim));
    recIFFT = real(my_ifft(measuredK));
end
toc

fprintf('GENFIRE: # of total volume: %d\n\n',size(measuredK(:)));
fprintf('GENFIRE: # of interpolated K: %d\n\n',sum(measuredK(:)~=0));

recIFFTsmall = recIFFT(ncK-n2:ncK+n2-1,ncK-n2:ncK+n2-1,ncK-n2:ncK+n2-1);%take back original sized array of the initial IFFT to compare before/after iteration


tic

if complexCutoff == 1;
    complexCutoff = 1.5; %% if all of the information is enforceable, I choose to also enforce super resolution as simulations 
    %% indicated doing so produced better reconstructions
end
constraintInd_complex = find(measuredK~=0&KMags<=complexCutoff);%indices of measured Fourier points to enforce with complex values
constraintInd_magnitude = find(measuredK~=0&KMags<=magnitudeCutoff);
% constraintInd_magnitude = [];%This is a null placeholder for future versions


% recIFFT=smooth3D(recIFFT,complexCutoff);%To low pass filter the IFFT reconstruction (which will be our initial
%%object if none was inputted) to the resolution that we can enforce.

% run GENFIRE iteration
fprintf('GENFIRE: Reconstructing... \n\n');
if isempty(initialObject) 
    [GENFIRE_rec, errK] = GENFIRE_iterate(numIterations,recIFFT,support,measuredK,constraintInd_complex,constraintInd_magnitude,[],[]);%run iterations
else
    [GENFIRE_rec, errK] = GENFIRE_iterate(numIterations,initialObject,support,measuredK,constraintInd_complex,constraintInd_magnitude,[],[]);
end
reconstructionTime = toc;
reconstructionTime = round(10*reconstructionTime)./10;
fprintf('GENFIRE: Reconstruction completed in %.12g seconds\n\n',reconstructionTime);

GENFIRE_rec = GENFIRE_rec(ncK-n2:ncK+n2-1,ncK-n2:ncK+n2-1,ncK-n2:ncK+n2-1);%extract the original sized result from the oversampled reconstruction


%display results
figure,
subplot(2,3,4), imagesc(squeeze(sum(GENFIRE_rec,1))),axis image,title('GENFIRE projection 1')
subplot(2,3,5), imagesc(squeeze(sum(GENFIRE_rec,2))),axis image,title('GENFIRE projection 2')
subplot(2,3,6), imagesc(squeeze(sum(GENFIRE_rec,3))),axis image,title('GENFIRE projection 3')
subplot(2,3,1), imagesc(squeeze(sum(recIFFTsmall,1))),axis image,title('before iteration projection 1')
subplot(2,3,2), imagesc(squeeze(sum(recIFFTsmall,2))),axis image,title('before iteration projection 2')
subplot(2,3,3), imagesc(squeeze(sum(recIFFTsmall,3))),axis image,title('before iteration projection 3')

nc = round((size(GENFIRE_rec,2)+1)/2);
figure,
subplot(2,3,4), imagesc(squeeze(GENFIRE_rec(nc,:,:))),axis image,title('GENFIRE slice 1')
subplot(2,3,5), imagesc(squeeze(GENFIRE_rec(:,nc,:))),axis image,title('GENFIRE slice 2')
subplot(2,3,6), imagesc(squeeze(GENFIRE_rec(:,:,nc))),axis image,title('GENFIRE slice 3')
subplot(2,3,1), imagesc(squeeze(recIFFTsmall(nc,:,:))),axis image,title('before iteration slice 1')
subplot(2,3,2), imagesc(squeeze(recIFFTsmall(:,nc,:))),axis image,title('before iteration slice 2')
subplot(2,3,3), imagesc(squeeze(recIFFTsmall(:,:,nc))),axis image,title('before iteration slice 3')



%save results
save(filename_results,'GENFIRE_rec','errK')







