%%  enforceableResolutionCutoffFinder %%

%%inputs:
%%  measuredK - measured reciprocal space points
%%  interpolationCutoffDistance - radius of sphere in which to include measured
%%      values when filling in grid. All points within this sphere will be weighted
%%      linearly by their inverse distance.
%%  Rfree_complex_target - For a given complex resolution enforcement cutoff, if Rfree complex 
%%      converges to a value less than or equal than Rfree_complex_target, then that resolution
%%      is considered enforceable
%%  Rfree_magnitude_target - same as Rfree_complex_target, but for magnitude enforcement range

%%outputs:
%%  complexCutoff - value between 0 and 1 indicating what percentage of the maximum resolution
%%       was found to be safely enforceable
%%  magnitudeCutoff - equivalent to complexCutoff except for magnitude enforcement
%%  Rfree_complex - value of complex Rfree vs iteration
%%  Rfree_magnitude - value of magnitude Rfree vs iteration

%% Author: AJ Pryor
%% Jianwei (John) Miao Coherent Imaging Group
%% University of California, Los Angeles
%% Copyright (c) 2015. All Rights Reserved.

function [complexCutoff, magnitudeCutoff, Rfree_complex, Rfree_magnitude] = enforceableResolutionCutoffFinder(measuredK,interpolationCutoffDistance,Rfree_complex_target,Rfree_magnitude_target)

global numIterations
global support
global pixelSize
percentValuesForRfree = 0.05;
terminationCondition = 0.005; % if the cutoff moves by less than this percentage
%then terminate the search

ncK = size(measuredK,2)/2+1;
n2K = ncK-1;
[Kx Ky Kz] = meshgrid(-n2K:n2K-1,-n2K:n2K-1,-n2K:n2K-1);
KMags = sqrt(Kx.^2+Ky.^2+Kz.^2);
KMags = KMags./n2K; %normalized Fourier indices

%begin binary search at half of maximum resolution
complexCutoff = 0.5;
complexCutoffFloor = 0;
complexCutoffCeil = 1;

resRange = -0.05; %thickness of highest resolution shell used for Rfree calculations

successfullyFoundComplexCutoff = 0;%flag to terminate while loop
successfullyFoundMagnitudeCutoff = 0;

fprintf('GENFIRE: Searching for maximum resolution to enforce complex values...\n\n')
while ~successfullyFoundComplexCutoff
    if exist('pixelSize','var')
    fprintf('GENFIRE: Trying %4.4g Angstrom cutoff (%.12g%% of maximum resolution)...\n\n',1/(round(100*complexCutoff)*2./pixelSize./100),round(10000*complexCutoff)./100)
    else
    fprintf('GENFIRE: Trying %.12g%% of maximum resolution...\n\n',round(10000*complexCutoff)./100)
    end
measuredPointInd_complex = find(measuredK~=0&KMags>=(complexCutoff+resRange)&KMags<complexCutoff); %candidate values for Rfree_complex
P = randperm(numel(measuredPointInd_complex)); %shuffle values
measuredPointInd_complex = measuredPointInd_complex(P);
cutoffInd_complex = floor(numel(measuredPointInd_complex).*percentValuesForRfree); %take 5% of values
R_freeInd_complex = measuredPointInd_complex(1:cutoffInd_complex);
initObject = zeros(size(measuredK),'single'); %start with all zero object to prevent bias in resolution cutoff search
acceptableComplexInd = find(measuredK~=0&KMags<=complexCutoff); %find all enforceable indices within resolution cutoff
constraintInd_complex = setdiff(acceptableComplexInd,R_freeInd_complex); %withhold the Rfree values from constraints
constraintInd_magnitude = [];%no magnitude constraints at this point
R_freeInd_mag = [];%no Rfree_mag at this point
[rec, err, Rfree_complex, Rfree_magnitude] = GENFIRE_iterate(numIterations,initObject,support,measuredK,constraintInd_complex,constraintInd_magnitude,R_freeInd_complex,R_freeInd_mag);
finalRfree_complex = Rfree_complex(end); %record final value of Rfree_complex
convergenceStatus = testConvergence(Rfree_complex);%check that the error converged 

%determine which way to move the cutoff. Each time it is moved by half of
%the distance between its current location and the closest point that it
%has already tested. This binary style of searching is much faster than
%just scanning.
if finalRfree_complex<Rfree_complex_target && convergenceStatus
    %%move cutoff to higher resolution
    distanceToMove = (complexCutoffCeil-complexCutoff)/2;
    if distanceToMove < terminationCondition %if the move is small enough, terminate the search
        successfullyFoundComplexCutoff = 1;%terminate loop by changing the flag
        complexCutoff = complexCutoffCeil; %choose the higher resolution of these two very close values
    else
        complexCutoffFloor = complexCutoff;%new floor is the current cutoff value
        complexCutoff = complexCutoff + distanceToMove;%update cutoff

    end
else
    %%move cutoff to lower resolution
    distanceToMove = (complexCutoffFloor-complexCutoff)/2;
    if abs(distanceToMove) < terminationCondition
        successfullyFoundComplexCutoff = 1;
        
    else
        complexCutoffCeil = complexCutoff;
        complexCutoff = complexCutoff + distanceToMove;

    end
end




end
tmpCC = round(10000*complexCutoff)./100;
    if exist('pixelSize','var')
        fprintf('GENFIRE: Convergence reached. Successfully found complex cutoff at %4.4g Angstrom cutoff (%.12g%% of maximum resolution)...\n\n',1/(round(100*complexCutoff)*2./pixelSize./100),round(10000*complexCutoff)./100)
    else
        fprintf('GENFIRE: Convergence reached. Successfully found complex cutoff at %.12g%% of maximum resolution\n\n',tmpCC)
    end
%%%%%%%For now the magnitude cutoff finder is being suppressed until a better
% %%%%%%%%%method is determined
% %%%%%%%%%%%
magnitudeCutoff = 0;
% %%%%%%%%%%%%
% 
% % initialObject = rec;
% initialObject = zeros(size(measuredK));
% magnitudeCutoff = complexCutoff;
% magnitudeCutoffFloor = complexCutoff;
% magnitudeCutoffCeil = 1;
% while ~successfullyFoundMagnitudeCutoff
% 
%     
% measuredPointInd_magnitude = find(measuredK~=0&KMags>=(magnitudeCutoff+resRange)&KMags<magnitudeCutoff);
% P = randperm(numel(measuredPointInd_magnitude));
% measuredPointInd_magnitude = measuredPointInd_magnitude(P);
% cutoffInd_magnitude = floor(numel(measuredPointInd_magnitude).*percentValuesForRfree);
% R_freeInd_magnitude = measuredPointInd_magnitude(1:cutoffInd_magnitude);
% 
% % % initObject = zeros(size(measuredK),'single');
% 
% acceptablemagnitudeInd = find(measuredK~=0&KMags<=magnitudeCutoff);%&KMags%>complexCutoff);
% constraintInd_magnitude = setdiff(acceptablemagnitudeInd,R_freeInd_magnitude);
% % constraintInd_complex = find(measuredK~=0&KMags<=complexCutoff);
% R_freeInd_complex = [];
% %             acceptableComplexInd = find(measuredK~=0&KMags<=magnitudeCutoff);
% %             constraintInd_complex = setdiff(acceptableComplexInd,R_freeInd_complex);
% % tmpK = my_fft(rec);
% % tmpK(R_freeInd_magnitude) = 0;
% % rec = real(my_ifft(tmpK));
% % initialObject(R_freeInd_magnitude) = 0;
% clear tmpK
% 
% [rec, err, Rfree_complex, Rfree_magnitude] = GENFIRE_iterate(numIterations,initialObject,support,measuredK,constraintInd_complex,constraintInd_magnitude,R_freeInd_complex,R_freeInd_magnitude);
% finalRfree_magnitude = Rfree_magnitude(end);
% 
% figure, plot(err),title(num2str(magnitudeCutoff))
% 
% % figure, plot(Rfree_magnitude),title(num2str(magnitudeCutoff))
% % figure, plot(R_jose),title(num2str(magnitudeCutoff))
% % mask = model>0;
% % R_real = sum(abs(model(mask)-rec(mask)))./sum(abs(model(mask)));
% % xlabel(num2str(R_real))
% % magnitudeCutoff
% 
% 
% convergenceStatus = testConvergence(Rfree_magnitude);
% if finalRfree_magnitude<Rfree_magnitude_target && convergenceStatus
%     %%move cutoff to higher resolution
%     distanceToMove = (magnitudeCutoffCeil-magnitudeCutoff)/2;
%     if distanceToMove < terminationCondition
%         successfullyFoundMagnitudeCutoff = 1;
%         magnitudeCutoff = magnitudeCutoffCeil; %choose the higher resolution of these two very close values
%     else
%         magnitudeCutoffFloor = magnitudeCutoff;
%         magnitudeCutoff = magnitudeCutoff + distanceToMove;
% 
%     end
% else
%     %%move cutoff to lower resolution
%     distanceToMove = (magnitudeCutoffFloor-magnitudeCutoff)/2;
%     if abs(distanceToMove) < terminationCondition
%         successfullyFoundMagnitudeCutoff = 1;
%         
%     else
%         magnitudeCutoffCeil = magnitudeCutoff;
%         magnitudeCutoff = magnitudeCutoff + distanceToMove;
% 
%     end
% end
% 
% end
% 
% tmpCC = round(complexCutoff*10000)/100;
% tmpMC = round(magnitudeCutoff*10000)/100;
% 
% fprintf('Successfully found complex cutoff at %d%% and magnitude cutoff at %d%%\n\n',tmpCC,tmpMC)




